@@ -5,9 +5,9 @@
 ;; Author: Yuan Fu <casouri@gmail.com>
 ;; Maintainer: Yuan Fu <casouri@gmail.com>
 ;; URL: https://github.com/casouri/vundo
-;; Version: 1.0.0
-;; Keywords: undo, text
-;; Package-Requires: ((emacs "28.0"))
+;; Version: 2.0.0
+;; Keywords: undo, text, editing
+;; Package-Requires: ((emacs "28.1"))
 ;;
 ;; This file is part of GNU Emacs.
 ;;
@@ -26,26 +26,39 @@
 
 ;;; Commentary:
 ;;
-;; To use vundo, type M-x vundo RET in the buffer you want to undo.
-;; A undo tree buffer should pop up. To move around, type:
+;; Vundo (visual undo) displays the undo history as a tree and lets you
+;; move in the tree to go back to previous buffer states. To use vundo,
+;; type M-x vundo RET in the buffer you want to undo. An undo tree buffer
+;; should pop up. To move around, type:
 ;;
 ;;   f   to go forward
 ;;   b   to go backward
+;;
 ;;   n   to go to the node below when you at a branching point
 ;;   p   to go to the node above
+;;
+;;   a   to go back to the last branching point
+;;   e   to go forward to the end/tip of the branch
+;;
 ;;   q   to quit, you can also type C-g
 ;;
-;; By default, you need to press RET to “commit” your change and if
-;; you quit with q or C-g, the change made by vundo are rolled back.
-;; You can set ‘vundo-roll-back-on-quit’ to nil to disable rolling
-;; back.
+;; n/p may need some more explanation. In the following tree, n/p can
+;; move between A and B because they share a parent (thus at a branching
+;; point), but not C and D.
 ;;
-;; If you bring up the vundo buffer and make some modification in the
-;; original buffer, the tree in the vundo buffer doesn’t automatically
-;; update. Vundo catches up the next time you invoke any command:
-;; instead of performing that command, it updates the tree.
+;;          A  C
+;;     ──○──○──○──○──○
+;;       │  ↕
+;;       └──○──○──○
+;;          B  D
 ;;
-;; Faces:
+;; By default, you need to press RET to “commit” your change and if you
+;; quit with q or C-g, the changes made by vundo are rolled back. You can
+;; set ‘vundo-roll-back-on-quit’ to nil to disable rolling back.
+;;
+;; Note: vundo.el requires Emacs 28.
+;;
+;; Customizable faces:
 ;;
 ;; - vundo-default
 ;; - vundo-node
@@ -70,6 +83,14 @@
 ;; to use that font:
 ;;
 ;;     (set-face-attribute 'vundo-default nil :family "Symbola")
+;;
+;; Comparing to undo-tree:
+;;
+;; Vundo doesn’t need to be turned on all the time nor replace the undo
+;; commands like undo-tree does. Vundo displays the tree horizontally,
+;; whereas undo-tree displays a tree vertically. Vundo doesn’t have many
+;; advanced features that undo-tree does (like showing diff), and most
+;; probably will not add those features in the future.
 
 ;;; Developer:
 ;;
@@ -93,10 +114,10 @@
 ;; `vundo--build-tree'. We build the tree by a simple observation:
 ;; only non-undo modifications creates new unique buffer states and
 ;; need to be drawn in the tree. For undo modifications, they
-;; associates equivalent nodes.
+;; associate equivalent nodes.
 ;;
 ;; Once we have generated the data structure and drawn the tree, vundo
-;; commands can move around on that tree by calling
+;; commands can move around in that tree by calling
 ;; `vundo--move-to-node'. It will construct the correct undo-list and
 ;; feed it to `primitive-undo'. `vundo--trim-undo-list' can trim the
 ;; undo list when possible.
@@ -114,14 +135,14 @@
 ;; Position-only records
 ;;
 ;; We know how undo works: when undoing, ‘primitive-undo’ looks at
-;; each record in ‘pending-undo-list’ and modify the buffer
+;; each record in ‘pending-undo-list’ and modifies the buffer
 ;; accordingly, and that modification itself pushes new undo records
 ;; into ‘buffer-undo-list’. However, not all undo records introduce
 ;; modification, if the record is an integer, ‘primitive-undo’ simply
 ;; ‘goto’ that position, which introduces no modification to the
 ;; buffer and pushes no undo record to ‘buffer-undo-list’. Normally
 ;; position records accompany other buffer-modifying records, but if a
-;; particular record consist of only position records, we have
+;; particular record consists of only position records, we have
 ;; trouble: after an undo step, ‘buffer-undo-list’ didn’t grow, as far
 ;; as vundo tree-folding algorithm is concerned, we didn’t move.
 ;; Assertions expecting to see new undo records in ‘buffer-undo-list’
@@ -162,7 +183,7 @@
   "If non-nil, vundo will roll back the change when it quits."
   :type 'boolean)
 
-(defcustom vundo--window-max-height 3
+(defcustom vundo-window-max-height 3
   "The maximum height of the vundo window."
   :type 'integer)
 
@@ -232,6 +253,19 @@ characters."
 		        :value-type (character :tag "Draw using")
 		        :options ,(mapcar #'car vundo-unicode-symbols)))
 
+(defcustom vundo-pre-enter-hook nil
+  "List of functions to call when entering vundo.
+This hook runs immediately after ‘vundo’ is called, in the buffer
+the user invoked ‘vundo’, before every setup ‘vundo’ does."
+  :type 'hook)
+
+(defcustom vundo-post-exit-hook nil
+  "List of functions to call when entering vundo.
+This hook runs in the original buffer the user invoked ‘vundo’,
+after every clean up the exiting function does. Ie, it is the
+very last thing that happens when vundo exists."
+  :type 'hook)
+
 ;;; Undo list to mod list
 
 (cl-defstruct vundo-m
@@ -342,18 +376,27 @@ modification in MOD-LIST. Return HASH-TABLE."
 ;; We know 3 and 5 are in the same equivalent list because 5 maps to 3
 ;; in `undo-equiv-table' (basically).
 
-(defun vundo--eqv-list-of (mod)
-  "Return all the modifications equivalent to MOD."
+(defun vundo--master-eqv-mod-of (mod)
+  "Return the master mod in the eqv-list of MOD.
+Master mod is the mod with the smallest index in the eqv-list.
+This function is equivalent to (car (vundo--eqv-list-of mod))."
   (while (vundo-m-prev-eqv mod)
     (cl-assert (not (eq mod (vundo-m-prev-eqv mod))))
     (setq mod (vundo-m-prev-eqv mod)))
-  ;; At the first mod in the equiv chain.
+  mod)
+
+(defun vundo--eqv-list-of (mod)
+  "Return all the modifications equivalent to MOD."
+  (while (vundo-m-next-eqv mod)
+    (cl-assert (not (eq mod (vundo-m-next-eqv mod))))
+    (setq mod (vundo-m-next-eqv mod)))
+  ;; Start at the last mod in the equiv chain, walk back to the first.
   (let ((eqv-list (list mod)))
-    (while (vundo-m-next-eqv mod)
-      (cl-assert (not (eq mod (vundo-m-next-eqv mod))))
-      (setq mod (vundo-m-next-eqv mod))
+    (while (vundo-m-prev-eqv mod)
+      (cl-assert (not (eq mod (vundo-m-prev-eqv mod))))
+      (setq mod (vundo-m-prev-eqv mod))
       (push mod eqv-list))
-    (nreverse eqv-list)))
+    eqv-list))
 
 (defun vundo--eqv-merge (mlist)
   "Connect modifications in MLIST to be in the same equivalence list.
@@ -418,8 +461,7 @@ If FROM non-nil, build from FORM-th modification in MOD-LIST."
            ;; the child.
            (unless (eq m 0)
              (let* ((m-1 (aref mod-list (1- m)))
-                    ;; TODO: may need to optimize.
-                    (min-eqv-mod (car (vundo--eqv-list-of m-1))))
+                    (min-eqv-mod (vundo--master-eqv-mod-of m-1)))
                (setf (vundo-m-parent mod) min-eqv-mod)
                (let ((children (vundo-m-children min-eqv-mod)))
                  ;; If everything goes right, we should never encounter
@@ -476,7 +518,8 @@ Translate according to ‘vundo-glyph-alist’."
   "Draw the tree in MOD-LIST in current buffer."
   (let* ((root (aref mod-list 0))
          (node-queue (list root))
-         (inhibit-read-only t))
+         (inhibit-read-only t)
+         (inhibit-modification-hooks t))
     (erase-buffer)
     (while node-queue
       (let* ((node (pop node-queue))
@@ -575,7 +618,8 @@ WINDOW is the window that was/is displaying the vundo buffer."
 
   ;; Disable evil-mode, as normal-mode
   ;; key bindings override the ones set by vundo.
-  (when (boundp 'evil-emacs-state-modes)
+  (when (and (boundp 'evil-emacs-state-modes)
+             (not (memq 'vundo-mode evil-emacs-state-modes)))
     (push 'vundo-mode evil-emacs-state-modes)))
 
 (defvar-local vundo--prev-mod-list nil
@@ -680,7 +724,7 @@ This function modifies ‘vundo--prev-mod-list’,
 
 (defun vundo--current-node (mod-list)
   "Return the currently highlighted node in MOD-LIST."
-  (car (vundo--eqv-list-of (aref mod-list (1- (length mod-list))))))
+  (vundo--master-eqv-mod-of (aref mod-list (1- (length mod-list)))))
 
 (defun vundo--highlight-node (node)
   "Highlight NODE as current node."
@@ -701,6 +745,7 @@ This function modifies ‘vundo--prev-mod-list’,
   (interactive)
   (when (not (consp buffer-undo-list))
     (user-error "There is no undo history"))
+  (run-hooks 'vundo-pre-enter-hook)
   (let ((vundo-buf (vundo-1 (current-buffer))))
     (select-window
      (display-buffer-in-side-window
@@ -709,7 +754,7 @@ This function modifies ‘vundo--prev-mod-list’,
         (window-height . 3))))
     (set-window-dedicated-p nil t)
     (let ((window-min-height 3))
-      (fit-window-to-buffer nil vundo--window-max-height))
+      (fit-window-to-buffer nil vundo-window-max-height))
     (goto-char
      (vundo-m-point
       (vundo--current-node vundo--prev-mod-list)))
@@ -774,14 +819,20 @@ Roll back changes if `vundo-roll-back-on-quit' is non-nil."
       vundo--orig-buffer vundo--prev-mod-list))
    (with-current-buffer vundo--orig-buffer
      (setq-local buffer-read-only nil))
-   (kill-buffer-and-window)))
+   (let ((orig-buffer vundo--orig-buffer))
+     (kill-buffer-and-window)
+     (with-current-buffer orig-buffer
+       (run-hooks 'vundo-post-exit-hook)))))
 
 (defun vundo-confirm ()
   "Confirm change and close vundo window."
   (interactive)
   (with-current-buffer vundo--orig-buffer
     (setq-local buffer-read-only nil))
-  (kill-buffer-and-window))
+  (let ((orig-buffer vundo--orig-buffer))
+    (kill-buffer-and-window)
+    (with-current-buffer orig-buffer
+      (run-hooks 'vundo-post-exit-hook))))
 
 ;;; Traverse undo tree
 
